import os
import json
import psycopg2
from psycopg2.extras import RealDictCursor
import fcntl
import tempfile

lockfile = 0
lockfilename = tempfile.gettempdir() + '/majisys_fews.lock'

def lock():
    global lockfile
    lockfile = open(lockfilename,'w')
    fcntl.flock(lockfile, fcntl.LOCK_EX)

def unlock():
    global lockfile
    fcntl.flock(lockfile, fcntl.LOCK_UN)
    lockfile.close()
    try:
        os.remove(lockfilename)
    except OSError:
        pass

def importdb(filename, db_params, parameters, locationset):
    ofile = open(filename, 'w')
    ofile.write(
'''<?xml version="1.0" encoding="UTF-8"?>
<timeSeriesImportRun xmlns="http://www.wldelft.nl/fews" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.wldelft.nl/fews http://fews.wldelft.nl/schemas/version1.0/timeSeriesImportRun.xsd">
	<import>
		<general>
			<importTypeStandard>database</importTypeStandard>
			<jdbcDriverClass>org.postgresql.Driver</jdbcDriverClass>
			<jdbcConnectionString>jdbc:postgresql://%s:%s/%s</jdbcConnectionString>
			<user>%s</user>
			<password>%s</password>
			<table name="monitoring.dump">
				<dateTimeColumn name="datetime" pattern="yyyy-MM-dd HH:mm:ss"/>
				<valueColumn name="scalarvalue"/>
				<parameterColumn name="parameterid"/>
				<locationColumn name="locationid"/>
				<!-- <unitColumn name="unit"/> -->
				<flagColumn name="flags"/>
				<commentColumn name="commenttext"/>
			</table>
			<missingValue>NaN</missingValue>
			<dataFeedId>Import_From_DB</dataFeedId>
		</general>
''' % (db_params['host'], db_params['port'], db_params['dbname'], db_params['user'], db_params['password']))
    for parameter in parameters:
        ofile.write(
'''		<timeSeriesSet>
			<moduleInstanceId>Import_From_DB</moduleInstanceId>
			<valueType>scalar</valueType>
			<parameterId>%s</parameterId>
			<locationSetId>%s</locationSetId>
			<timeSeriesType>external historical</timeSeriesType>
			<timeStep unit="nonequidistant"/>
			<readWriteMode>add originals</readWriteMode>
			<synchLevel>1</synchLevel>
		</timeSeriesSet>
''' % (parameter['id'], locationset))
    ofile.write(
'''	</import>
</timeSeriesImportRun>''')
    ofile.flush()
    ofile.close()

def clientconfig(filename, db_params):
    ofile = open(filename, 'w')
    ofile.write(
'''<?xml version="1.0" encoding="UTF-8"?>
<clientConfiguration xmlns="http://www.wldelft.nl/fews" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.wldelft.nl/fews http://fews.wldelft.nl/schemas/version1.0/clientConfig.xsd">
	<localDataStoreFormat>Firebird</localDataStoreFormat>
	<externalTables>
		<databaseServer>
			<dbServerType>postgresql</dbServerType>
			<dbServerName>%s</dbServerName>
			<dbServerPort>%s</dbServerPort>
			<dbInstanceName>%s</dbInstanceName>
			<dbInstanceUser>%s</dbInstanceUser>
			<dbInstancePassword>%s</dbInstancePassword>
		</databaseServer>
		<parameterGroupsTable>
			<table>fews.parametergroups</table>
			<keyColumn>groupKey</keyColumn>
			<idColumn>id</idColumn>
			<nameColumn>name</nameColumn>
			<descriptionColumn>description</descriptionColumn>
			<parameterTypeColumn>parameterType</parameterTypeColumn>
			<unitColumn>unit</unitColumn>
			<displayUnitColumn>displayUnit</displayUnitColumn>
		</parameterGroupsTable>
		<parametersTable>
			<table>fews.parameterstable</table>
			<keyColumn>parameterKey</keyColumn>
			<idColumn>id</idColumn>
			<nameColumn>name</nameColumn>
			<shortNameColumn>shortName</shortNameColumn>
			<descriptionColumn>description</descriptionColumn>
			<parameterGroupKeyColumn>groupKey</parameterGroupKeyColumn>
			<valueResolutionColumn>valueResolution</valueResolutionColumn>
		</parametersTable>
		<locationsTable>
			<geoDatum>WGS 1984</geoDatum>
			<table>fews.wgslocations</table>
			<keyColumn>locationKey</keyColumn>
			<idColumn>id</idColumn>
			<nameColumn>name</nameColumn>
			<shortNameColumn>shortName</shortNameColumn>
			<descriptionColumn>description</descriptionColumn>
			<iconColumn>icon</iconColumn>
			<toolTipColumn>toolTip</toolTipColumn>
			<parentLocationIdColumn>parentLocationId</parentLocationIdColumn>
			<visibilityStartTimeColumn>visibilityStartTime</visibilityStartTimeColumn>
			<visibilityEndTimeColumn>visibilityEndTime</visibilityEndTimeColumn>
			<xColumn>x</xColumn>
			<yColumn>y</yColumn>
			<zColumn>z</zColumn>
		</locationsTable>
		<qualifiersTable>
			<table>fews.qualifiers</table>
			<keyColumn>qualifierKey</keyColumn>
			<idColumn>id</idColumn>
			<descriptionColumn>description</descriptionColumn>
		</qualifiersTable>
		<qualifierSetsTable>
			<table>fews.qualifiersets</table>
			<keyColumn>qualifierSetKey</keyColumn>
			<idColumn>id</idColumn>
			<qualifierKey1Column>qualifierKey1</qualifierKey1Column>
			<qualifierKey2Column>qualifierKey2</qualifierKey2Column>
			<qualifierKey3Column>qualifierKey3</qualifierKey3Column>
			<qualifierKey4Column>qualifierKey4</qualifierKey4Column>
		</qualifierSetsTable>
		<moduleInstancesTable>
			<table>fews.moduleinstances</table>
			<keyColumn>moduleInstanceKey</keyColumn>
			<idColumn>id</idColumn>
			<nameColumn>name</nameColumn>
			<descriptionColumn>description</descriptionColumn>
		</moduleInstancesTable>
		<timeStepsTable>
			<table>fews.timesteps</table>
			<keyColumn>timeStepKey</keyColumn>
			<idColumn>id</idColumn>
			<labelColumn>label</labelColumn>
		</timeStepsTable>
		<aggregationPeriodsTable>
			<table>fews.aggregationperiods</table>
			<keyColumn>aggregationPeriodKey</keyColumn>
			<idColumn>id</idColumn>
		</aggregationPeriodsTable>
		<usersTableTable>
			<table>fews.users</table>
			<keyColumn>userKey</keyColumn>
			<idColumn>id</idColumn>
			<nameColumn>name</nameColumn>
		</usersTableTable>
		<timeSeriesKeysTable>
			<table>fews.timeserieskeys</table>
			<keyColumn>seriesKey</keyColumn>
			<parameterKeyColumn>parameterKey</parameterKeyColumn>
			<locationKeyColumn>locationKey</locationKeyColumn>
			<qualifierSetKeyColumn>qualifierSetKey</qualifierSetKeyColumn>
			<timeStepKeyColumn>timeStepKey</timeStepKeyColumn>
			<aggregationPeriodKeyColumn>aggregationPeriodKey</aggregationPeriodKeyColumn>
			<moduleInstanceKeyColumn>moduleInstanceKey</moduleInstanceKeyColumn>
			<valueTypeColumn>valuetype</valueTypeColumn>
			<modificationTimeColumn>modificationtime</modificationTimeColumn>
		</timeSeriesKeysTable>
		<timeSeriesValuesAndFlagsTable>
			<table>fews.timeSeriesvaluesandflags</table>
			<seriesKeyColumn>seriesKey</seriesKeyColumn>
			<timeStampColumn>dateTime</timeStampColumn>
			<valueColumn>scalarValue</valueColumn>
			<flagsColumn>flags</flagsColumn>
		</timeSeriesValuesAndFlagsTable>
		<timeSeriesCommentsTable>
			<table>fews.timeseriescomments</table>
			<seriesKeyColumn>seriesKey</seriesKeyColumn>
			<timeStampColumn>dateTime</timeStampColumn>
			<commentColumn>commentText</commentColumn>
		</timeSeriesCommentsTable>
		<timeSeriesManualEditsHistoryTable>
			<table>fews.timeseriesmanualeditshistory</table>
			<seriesKeyColumn>seriesKey</seriesKeyColumn>
			<timeStampColumn>dateTime</timeStampColumn>
			<editTimeStampColumn>editDateTime</editTimeStampColumn>
			<userKeyColumn>userKey</userKeyColumn>
			<valueColumn>scalarValue</valueColumn>
			<flagsColumn>flags</flagsColumn>
			<commentColumn>commentText</commentColumn>
		</timeSeriesManualEditsHistoryTable>
	</externalTables>
</clientConfiguration>
''' % (db_params['host'], db_params['port'], db_params['dbname'], db_params['user'], db_params['password']))
    ofile.flush()
    ofile.close()

def filters(filename, parameters, locationset):
    ofile = open(filename, 'w')
    ofile.write(
'''<?xml version="1.0" encoding="UTF-8"?>
<filters version="1.1" xmlns="http://www.wldelft.nl/fews" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.wldelft.nl/fews http://fews.wldelft.nl/schemas/version1.0/filters.xsd">
	<filter id="All obs (All FEWS)" name="All obs (All FEWS)">
		<child foreignKey="CSV"/>
		<child foreignKey="Database"/>
	</filter>
	<filter id="CSV" name="CSV">
''')
    for parameter in parameters:
        ofile.write(
'''		<timeSeriesSet>
			<moduleInstanceId>Import_CSV</moduleInstanceId>
			<valueType>scalar</valueType>
			<parameterId>%s</parameterId>
			<locationSetId>%s</locationSetId>
			<timeSeriesType>external historical</timeSeriesType>
			<timeStep unit="nonequidistant"/>
			<readWriteMode>add originals</readWriteMode>
		</timeSeriesSet>
''' % (parameter['id'], locationset))
    ofile.write('''	</filter>
	<filter id="Database" name="Database">
''')
    for parameter in parameters:
        ofile.write(
'''		<timeSeriesSet>
			<moduleInstanceId>Import_From_DB</moduleInstanceId>
			<valueType>scalar</valueType>
			<parameterId>%s</parameterId>
			<locationSetId>%s</locationSetId>
			<timeSeriesType>external historical</timeSeriesType>
			<timeStep unit="nonequidistant"/>
			<readWriteMode>add originals</readWriteMode>
			<synchLevel>1</synchLevel>
		</timeSeriesSet>
''' % (parameter['id'], locationset))
    ofile.write('''	</filter>
</filters>''')
    ofile.flush()
    ofile.close()

def locationsets(filename, locationset):
    ofile = open(filename, 'w')
    ofile.write(
'''<?xml version="1.0" encoding="UTF-8"?>
<locationSets version="1.1" xmlns="http://www.wldelft.nl/fews" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.wldelft.nl/fews http://fews.wldelft.nl/schemas/version1.0/locationSets.xsd">
	<locationSet id="%s"/>
</locationSets>
''' % (locationset))
    ofile.flush()
    ofile.close()

def importcsv(filename, parameters, locationset):
    ofile = open(filename, 'w')
    ofile.write(
'''<?xml version="1.0" encoding="UTF-8"?>
<timeSeriesImportRun xmlns="http://www.wldelft.nl/fews" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.wldelft.nl/fews http://fews.wldelft.nl/schemas/version1.0/timeSeriesImportRun.xsd">
	<!--Import data-->
	<import>
		<general>
			<importType>CSV</importType>
			<folder>$IMPORT_FOLDER$/CSV</folder>
			<failedFolder>$IMPORT_FAILED_FOLDER$/CSV</failedFolder>
			<backupFolder>$IMPORT_BACKUP_FOLDER$/CSV</backupFolder>
			<idMapId>IdImportCSV</idMapId>
			<missingValue>-999.0</missingValue>
			<importTimeZone>
				<timeZoneOffset>+00:00</timeZoneOffset>
			</importTimeZone>
		</general>
''')
    for parameter in parameters:
        ofile.write(
'''		<timeSeriesSet>
			<moduleInstanceId>Import_CSV</moduleInstanceId>
			<valueType>scalar</valueType>
			<parameterId>%s</parameterId>
			<locationSetId>%s</locationSetId>
			<timeSeriesType>external historical</timeSeriesType>
			<timeStep unit="nonequidistant"/>
			<readWriteMode>add originals</readWriteMode>
			<synchLevel>1</synchLevel>
		</timeSeriesSet>
''' % (parameter['id'], locationset))
    ofile.write('''	</import>
</timeSeriesImportRun>''')
    ofile.flush()
    ofile.close()

def importlakelevel(filename, locationset):
    ofile = open(filename, 'w')
    ofile.write(
'''<?xml version="1.0" encoding="UTF-8"?>
<timeSeriesImportRun xmlns="http://www.wldelft.nl/fews" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.wldelft.nl/fews http://fews.wldelft.nl/schemas/version1.0/timeSeriesImportRun.xsd">
	<!--Import data-->
	<import>
		<general>
			<importType>CSV</importType>
			<folder>$IMPORT_FOLDER$/CSV</folder>
			<failedFolder>$IMPORT_FAILED_FOLDER$/CSV</failedFolder>
			<backupFolder>$IMPORT_BACKUP_FOLDER$/CSV</backupFolder>
			<idMapId>IdImportWaterlevel</idMapId>
			<missingValue>NaN</missingValue>
			<importTimeZone>
				<timeZoneOffset>+00:00</timeZoneOffset>
			</importTimeZone>
		</general>
		<timeSeriesSet>
			<moduleInstanceId>Import_LakeLevel_CSV</moduleInstanceId>
			<valueType>scalar</valueType>
			<parameterId>L.obs</parameterId>
			<locationSetId>%s</locationSetId>
			<timeSeriesType>external historical</timeSeriesType>
			<timeStep unit="day" multiplier="1"/>
			<readWriteMode>add originals</readWriteMode>
			<synchLevel>1</synchLevel>
		</timeSeriesSet>
	</import>
</timeSeriesImportRun>
''' % (locationset))
    ofile.flush()
    ofile.close()

def log4jconfig(filename):
    ofile = open(filename, 'w')
    ofile.write(
'''<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE log4j:configuration SYSTEM "log4j.dtd">

<log4j:configuration xmlns:log4j="http://jakarta.apache.org/log4j/">
    <appender name="systemMonitorSynchPanel" class="nl.wldelft.fews.gui.plugin.systemmonitor.SynchLogPanelLogAppender">
        <param name="level" value="DEBUG"/>
    </appender>

    <appender name="explorerLogPanel" class="nl.wldelft.fews.gui.explorer.ExplorerLoggingPanelLogAppender">
        <param name="level" value="INFO"/>
    </appender>
    
    <appender name="scenarioEditorLogPanel" class="nl.wldelft.fews.gui.explorer.ExplorerLoggingPanelLogAppender">
        <param name="level" value="INFO"/>
    </appender>

    <appender name="explorerAcknowladgeIcon" class="nl.wldelft.fews.gui.explorer.FewsExplorerAcknowladgeIconLogAppender">
        <param name="level" value="INFO"/>
    </appender>

    <appender name="dataStoreLogEntriesTable" class="nl.wldelft.fews.util.LogEntriesTableLogAppender">
        <param name="level" value="INFO"/>
    </appender>

    <appender name="splashScreen" class="nl.wldelft.fews.gui.explorer.LoggingSplashScreenLogAppender">
        <param name="level" value="INFO"/>
    </appender>

    <appender name="workflowNavigatorLogPanel" class="nl.wldelft.fews.gui.plugin.workflownavigator.WorkflowNavigatorLogAppender">
                <param name="level" value="DEBUG"/>
    </appender>

<!--
    <appender name="defaultLogFile" class="org.apache.log4j.FileAppender">
        <param name="File" value="${log.file.path}"/>
        <param name="Append" value="false"/>
-->

    <appender name="defaultLogFile" class="org.apache.log4j.ConsoleAppender">
        <layout class="org.apache.log4j.PatternLayout">
            <!-- [Date] Priority - category - Message -->
            <param name="ConversionPattern" value="[%d] %p - %C{1}.%M - %m%n"/>
        </layout>

    </appender>

    <category name="nl.wldelft.fews.system.synch" additivity="true">
        <priority value="INFO"/>
        <appender-ref ref="systemMonitorSynchPanel"/>
        <appender-ref ref="explorerAcknowladgeIcon"/>
    </category>

    <category name="nl.wldelft.fews.system.synch.SynchTask" additivity="false">
        <priority value="INFO"/>
        <appender-ref ref="explorerLogPanel"/>
        <appender-ref ref="defaultLogFile"/>
    </category>

    <category name="nl.wldelft.fews.system.synch.SynchProfile" additivity="false">
        <priority value="ERROR"/>
        <appender-ref ref="explorerLogPanel"/>
        <appender-ref ref="defaultLogFile"/>
    </category>
    
    <category name="nl.wldelft.fews.gui.plugin.scenarioeditor" additivity="false">
        <priority value="INFO"/>
        <appender-ref ref="scenarioEditorLogPanel"/>
        <appender-ref ref="defaultLogFile"/>
    </category>

    <category name="nl.wldelft.fews.gui.plugin.workflownavigator" additivity="false">
                <priority value="DEBUG"/>
                <appender-ref ref="workflowNavigatorLogPanel"/>
                <appender-ref ref="defaultLogFile"/>
    </category>

    <category name="ucar" additivity="false">
        <priority value="ERROR"/>
        <appender-ref ref="explorerLogPanel"/>
        <appender-ref ref="defaultLogFile"/>
    </category>
  
    <category name="org.jboss" additivity="false">
        <priority value="INFO"/>
        <appender-ref ref="defaultLogFile"/>
    </category>

    <category name="org.springframework" additivity="false">
        <priority value="WARN"/>
        <appender-ref ref="explorerLogPanel"/>
        <appender-ref ref="defaultLogFile"/>
        <appender-ref ref="systemMonitorSynchPanel"/>
    </category>

    <category name="org.mortbay" additivity="false">
        <priority value="WARN"/>
        <appender-ref ref="explorerLogPanel"/>
        <appender-ref ref="defaultLogFile"/>
        <appender-ref ref="systemMonitorSynchPanel"/>
    </category>
    
    <root>
        <priority value="INFO"/>
        <appender-ref ref="defaultLogFile"/>
        <appender-ref ref="splashScreen"/>
        <appender-ref ref="explorerLogPanel"/>
        <appender-ref ref="explorerAcknowladgeIcon"/>
        <appender-ref ref="dataStoreLogEntriesTable"/>
    </root>
</log4j:configuration>
''')
    ofile.flush()
    ofile.close()

def updatefewsxml():
    curpath = os.path.dirname(os.path.realpath( __file__ )) + '/'
    fewspath = '../../../../processing/fews/'

    import_db = curpath + fewspath + 'FEWS_Naivasha/Config/ModuleConfigFiles/Import/Import_From_DB.xml'
    fews_filters = curpath + fewspath + 'FEWS_Naivasha/Config/RegionConfigFiles/Filters.xml'
    client_config = curpath + fewspath + 'FEWS_Naivasha/clientConfig.xml'
    location_sets = curpath + fewspath + 'FEWS_Naivasha/Config/RegionConfigFiles/LocationSets.xml'
    import_csv = curpath + fewspath + 'FEWS_Naivasha/Config/ModuleConfigFiles/Import/Import_CSV.xml'
    import_lakelevel = curpath + fewspath + 'FEWS_Naivasha/Config/ModuleConfigFiles/Import/Import_LakeLevel_CSV.xml'
    log4j_config = curpath + fewspath + 'FEWS_Naivasha/Log4jConfig.xml'

    file = open("../.credentials/.postgres")
    connection_string = file.readline()
    pg = psycopg2.connect(connection_string)
    query = pg.cursor(cursor_factory=RealDictCursor)

    file = open("../.credentials/.postgresadmin")
    connection_string = file.readline()
    connection_string = '{"' + connection_string.replace(' ',',"').replace('=','":').replace("'",'"') + '}'
    db_params = json.loads(connection_string)
    # 'host', 'port', 'dbname', 'user', 'password'

    locationset = 'All_Locations'

    sql = "SELECT DISTINCT id FROM fews.parameterstable ORDER BY id;"
    query.execute(sql)
    parameters= query.fetchall()
    
    lock()
    importdb(import_db, db_params, parameters, locationset)
    filters(fews_filters, parameters, locationset)
    clientconfig(client_config, db_params)
    locationsets(location_sets, locationset)
    importcsv(import_csv, parameters, locationset)
    importlakelevel(import_lakelevel, locationset)
    log4jconfig(log4j_config)
    unlock()

def probeaccess():
    curpath = os.path.dirname(os.path.realpath( __file__ )) + '/'
    fewspath = '../../../../processing/fews/'

    import_db = curpath + fewspath + 'FEWS_Naivasha/Config/ModuleConfigFiles/Import/Import_From_DB.xml'
    fews_filters = curpath + fewspath + 'FEWS_Naivasha/Config/RegionConfigFiles/Filters.xml'
    client_config = curpath + fewspath + 'FEWS_Naivasha/clientConfig.xml'
    location_sets = curpath + fewspath + 'FEWS_Naivasha/Config/RegionConfigFiles/LocationSets.xml'
    import_csv = curpath + fewspath + 'FEWS_Naivasha/Config/ModuleConfigFiles/Import/Import_CSV.xml'
    import_lakelevel = curpath + fewspath + 'FEWS_Naivasha/Config/ModuleConfigFiles/Import/Import_LakeLevel_CSV.xml'
    log4j_config = curpath + fewspath + 'FEWS_Naivasha/Log4jConfig.xml'
    ofile = open(import_db, 'w')
    ofile.close()
    ofile = open(fews_filters, 'w')
    ofile.close()
    ofile = open(client_config, 'w')
    ofile.close()
    ofile = open(location_sets, 'w')
    ofile.close()
    ofile = open(import_csv, 'w')
    ofile.close()
    ofile = open(import_lakelevel, 'w')
    ofile.close()
    ofile = open(log4j_config, 'w')
    ofile.close()
