#!/usr/bin/python3

import os
import cgi
import json
import re
from datetime import datetime
import struct
import psycopg2
from psycopg2.extras import RealDictCursor
from importlib.machinery import SourceFileLoader
import fewsdb

# time is a string in dateformat, it will be converted to postgres ISO time string (YYYY-MM-DD HH:MM:SS)
def dateconvertT(time, dateformat): # dateformat: dmy, ymd, mdy, ydm
    time = time.replace('/','-').strip() # normalize
    twodigityear = False
    if dateformat == 'dmy':
        try:
            time = datetime.strptime(time, "%d-%m-%Y %H:%M:%S")
        except ValueError:
            try:
                time = datetime.strptime(time, "%d-%m-%Y %H:%M")
            except ValueError:
                try:
                    time = datetime.strptime(time, "%d-%m-%Y")
                except ValueError:
                    twodigityear = True
                    try:
                        time = datetime.strptime(time, "%d-%m-%y %H:%M:%S")
                    except ValueError:
                        try:
                            time = datetime.strptime(time, "%d-%m-%y %H:%M")
                        except ValueError:
                            time = datetime.strptime(time, "%d-%m-%y")
    elif dateformat == 'ymd': # probably obsolete, but we have it here to supplement the seconds if they are missing
        try:
            time = datetime.strptime(time, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            try:
                time = datetime.strptime(time, "%Y-%m-%d %H:%M")
            except ValueError:
                try:
                    time = datetime.strptime(time, "%Y-%m-%d")
                except ValueError:
                    twodigityear = True
                    try:
                        time = datetime.strptime(time, "%y-%m-%d %H:%M:%S")
                    except ValueError:
                        try:
                            time = datetime.strptime(time, "%y-%m-%d %H:%M")
                        except ValueError:
                            time = datetime.strptime(time, "%y-%m-%d")
    elif dateformat == 'mdy':
        try:
            time = datetime.strptime(time, "%m-%d-%Y %H:%M:%S")
        except ValueError:
            try:
                time = datetime.strptime(time, "%m-%d-%Y %H:%M")
            except ValueError:
                try:
                    time = datetime.strptime(time, "%m-%d-%Y")
                except ValueError:
                    twodigityear = True
                    try:
                        time = datetime.strptime(time, "%m-%d-%y %H:%M:%S")
                    except ValueError:
                        try:
                            time = datetime.strptime(time, "%m-%d-%y %H:%M")
                        except ValueError:
                            time = datetime.strptime(time, "%m-%d-%y")
    elif dateformat == 'ydm':
        try:
            time = datetime.strptime(time, "%Y-%d-%m %H:%M:%S")
        except ValueError:
            try:
                time = datetime.strptime(time, "%Y-%d-%m %H:%M")
            except ValueError:
                try:
                    time = datetime.strptime(time, "%Y-%d-%m")
                except ValueError:
                    twodigityear = True
                    try:
                        time = datetime.strptime(time, "%y-%d-%m %H:%M:%S")
                    except ValueError:
                        try:
                            time = datetime.strptime(time, "%y-%d-%m %H:%M")
                        except ValueError:
                            time = datetime.strptime(time, "%y-%d-%m")
    if twodigityear and (time.year > datetime.now().year):
        time = datetime(time.year - 100, time.month, time.day, time.hour, time.minute, time.second)
    time = time.strftime("%Y-%m-%d %H:%M:%S") # get postgres ISO string
    return time

# date is a string in dateformat, time is HH:MM or HH:MM:SS, date + time will be converted to postgres ISO time string (YYYY-MM-DD HH:MM:SS)
def dateconvertDT(date, time, dateformat): # dateformat: dmy, ymd, mdy, ydm
    return dateconvertT(date + " " + time, dateformat)

def ReadDouble(ifile):
    buffer = ifile.read(8)
    rVal = struct.unpack(">d", buffer)[0]
    return rVal

def ReadFloat(ifile):
    buffer = ifile.read(4)
    rVal = struct.unpack(">f", buffer)[0]
    return rVal

def ReadInt(ifile):
    buffer = ifile.read(4)
    iVal = struct.unpack(">i", buffer)[0]
    return iVal

def ReadUnsignedInt(ifile):
    buffer = ifile.read(4)
    iVal = struct.unpack(">I", buffer)[0]
    return iVal

def ReadShort(ifile):
    buffer = ifile.read(2)
    iVal = struct.unpack(">h", buffer)[0]
    return iVal

def ReadUnsignedShort(ifile):
    buffer = ifile.read(2)
    iVal = struct.unpack(">H", buffer)[0]
    return iVal

def ReadByte(ifile):
    buffer = ifile.read(1)
    b = struct.unpack("B", buffer)[0]
    return b
    
def ReadString(ifile):
    string = ifile.readline().decode('utf-8').rstrip()
    return string
    
def ReadMessage(ifile, nrbytes):
    buffer = ifile.read(nrbytes)
    string = ""
    for c in buffer:
        string += chr(c & 127)
    return string

def getSection(str, substr):
    start = len(substr) + str.find(substr)
    end = str.find(' ', start)
    str = str[start:end]
    return str

def sliceTimes(clk):
    res = []
    while clk != '':
        res.append(clk[0:8])
        clk = clk[8:]
    return list(reversed(res))

def sliceData(data, size):
    res = []
    while data != '':
        res.append(str(float(data[0:size])))
        data = data[size:]
    return list(reversed(res))

def fillTimes(clk, date, utc): # TODO: do this with timezone and offset
    res = []
    for time in clk:
        dtime = dateconvertDT(date, time, "dmy")
        dtime = datetime.strptime(dtime, "%Y-%m-%d %H:%M:%S") # to datetime object format
        if abs((dtime - utc).total_seconds()) > 43200: # 12 hours
            dtime = datetime(dtime.year, dtime.month, dtime.day + 1, dtime.hour, dtime.minute, dtime.second)
        dtime = dtime.strftime("%Y-%m-%d %H:%M:%S") # postgres ISO string
        res.append(dtime)
        return res

def addToValues(values, parameterid, times, data):
    if not parameterid in values:
        values[parameterid] = {}
        values[parameterid]['times'] = []
        values[parameterid]['values'] = []
    for i in range(0, len(times)):
        values[parameterid]['times'].append(times[i])
        values[parameterid]['values'].append(data[i])

def parseMessage(dcpmessage, line, values):
    start = 5 + line.find("0 at ")
    end = line.find(' ', start)
    date = line[start:end]
    end = line.find(' ', end + 1) # 2nd space
    line = line[start:end]
    utc = dateconvertT(line, "dmy")
    utc = datetime.strptime(utc, "%Y-%m-%d %H:%M:%S") # in datetime object format

    clk = getSection(dcpmessage, "CLK: ")
    bt = getSection(dcpmessage, "BT: ")
    rain = getSection(dcpmessage, "RAIN: ") # P.obs
    temp = getSection(dcpmessage, "TEMP: ") # T.obs
    hum = getSection(dcpmessage, "HUM: ") # RH.obs
    lvl = getSection(dcpmessage, "LVL: ") # H.obs
    wtmp = getSection(dcpmessage, "WTMP: ") # Ta.obs
    
    clk = sliceTimes(clk)
    clk = fillTimes(clk, date, utc)
    bt = sliceData(bt, 5)
    rain = sliceData(rain, 6) # P.obs
    temp = sliceData(temp, 6) # T.obs
    hum = sliceData(hum, 6) # RH.obs
    lvl = sliceData(lvl, 6) # H.obs
    wtmp = sliceData(wtmp, 6) # Ta.obs
    
    addToValues(values, 'P.obs', clk, rain)
    addToValues(values, 'T.obs', clk, temp)
    addToValues(values, 'RH.obs', clk, hum)
    addToValues(values, 'H.obs', clk, lvl)
    addToValues(values, 'Ta.obs', clk, wtmp)

def readdcp(ifile):
    count = 0
    values = {}
    line = ReadString(ifile)
    while line != '':
        count = count + 1
        line = ReadString(ifile)
        SubSystemId = ReadByte(ifile)
        ModuleId = ReadByte(ifile)
        ReceiverId = ReadByte(ifile)
        ChannelType = ReadByte(ifile)
        ChannelFreq = ReadUnsignedInt(ifile)
        FrequencyOffset = ReadInt(ifile)
        CarrierLevel = ReadFloat(ifile)
        ModulationLevel = ReadFloat(ifile)
        FrameSyncErrors = ReadByte(ifile)
        AddressErrors = ReadByte(ifile)
        EOTErrors = ReadByte(ifile)
        booldata = ReadByte(ifile)
        MessageLengthFlag = bool(booldata & 128)
        EOTCorrect = bool(booldata & 64)
        MessageDecodeFlag = bool(booldata & 32)
        TimeFrameAlarm = bool(booldata & 16)
        AbortReason = ReadByte(ifile)
        FrameSyncTime = (((ReadShort(ifile) << 16) + ReadShort(ifile)) << 16) + ReadShort(ifile)
        MessageLength = ReadUnsignedShort(ifile)
        DCPAddress = ReadUnsignedInt(ifile)
        DCPMessage = ReadMessage(ifile, MessageLength - 2 * 4)
        parseMessage(DCPMessage, line, values)
        DCPEOT = ReadUnsignedInt(ifile)
        line = ReadString(ifile)
    return count,values

# main()

form = cgi.FieldStorage()
uploadedFile = form['filedata-button']
location = form.getvalue('locationid-inputEl')
ticket = form.getvalue('ticket')
curpath = os.path.dirname(os.path.realpath( __file__ ))
cas = SourceFileLoader("cas", curpath + "/../management/cas.py").load_module()
if not cas.authorized(ticket,{"timeseries" : "rw"}):
    messages = ["Not authorized."]
else:
    username = cas.getusername(ticket)
    file = open("../.credentials/.postgres")
    connection_string = file.readline()
    pg = psycopg2.connect(connection_string)
    query = pg.cursor(cursor_factory=RealDictCursor)
    sql = "WITH spatialfilter AS \
           (SELECT ST_Union(geom) AS aoi FROM app.user JOIN app.user_spatialfilter USING(userid) JOIN app.spatialfilter USING (gid) \
            WHERE username='" + username + "') \
           SELECT station_id FROM fews.stations JOIN spatialfilter ON ST_Intersects(geom, spatialfilter.aoi) \
           WHERE station_id ='" + location + "'"
    query.execute(sql)
    if query.rowcount == 0:
        messages = ["Not authorized."]
    else:
        filename = uploadedFile.filename
        comment = "Imported from \"" + filename + "\""
        comment = comment[:64]
        messages = []
        try:
            dcpfile = uploadedFile.file
            count,values = readdcp(dcpfile)
            if count > 0: # no issues with the parsing
                comments = [comment]
                flags = []
                for parameter in values.keys():
                    times = values[parameter]['times']
                    vals = values[parameter]['values']
                    messages = fewsdb.importdb(parameter, location, times, vals, flags, comments)
            else:
                messages += ["Timeseries from DCP not inserted: Couldn't open DCP file. Is the file unzipped?"]
        except Exception as err:
            messages.append(str(err) + ".")
            messages.append("Timeseries from DCP not inserted: Couldn't open DCP file. Is the file unzipped?")

print("Content-type: application/json")
print()
print("{\"success\": true, \"messages\": " + json.dumps(messages) + "}")
